/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Quickfile extension for Thunderbird.
 *
 * The Initial Developer of the Original Code is Paul Tomlin.
 * Portions created by the Initial Developer are Copyright (C) 2005, 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Paul Tomlin
 *  Mark Shuttleworth
 *
 * ***** END LICENSE BLOCK ***** */

/* $Id: qfMailNewsFoldersAutoComplete.js,v 1.4 2006/11/24 07:44:21 paultomlin Exp $ */

// UUID uniquely identifying our components
// From: http://kruithof.xs4all.nl/uuid/uuidgen
const XPFE_CLASS_ID    = Components.ID("{7da86ab0-5d39-11db-b0de-0800200c9a66}");
const TOOLKIT_CLASS_ID = Components.ID("{c0cc15b0-5dd1-11db-b0de-0800200c9a66}");

// description
const XPFE_CLASS_NAME = "Quickfile Mail News Folders Auto Complete Session XPCOM Component";
const TOOLKIT_CLASS_NAME = "Quickfile Mail News Folders Auto Complete Search XPCOM Component";

// textual unique identifier
const XPFE_CONTRACT_ID = "@mozilla.org/autocompleteSession;1?type=mailnewsfolders";
const TOOLKIT_CONTRACT_ID = "@mozilla.org/autocomplete/search;1?name=mailnewsfolders";

// static import interfaces
const nsIAutoCompleteItem               = Components.interfaces.nsIAutoCompleteItem;
const nsIAutoCompleteResult             = Components.interfaces.nsIAutoCompleteResult;
const nsIAutoCompleteResults            = Components.interfaces.nsIAutoCompleteResults;
const nsIAutoCompleteSearch             = Components.interfaces.nsIAutoCompleteSearch;
const nsIAutoCompleteSession            = Components.interfaces.nsIAutoCompleteSession;
const nsIAutoCompleteStatus             = Components.interfaces.nsIAutoCompleteStatus;
const nsICategoryManager                = Components.interfaces.nsICategoryManager;
const nsIComponentRegistrar             = Components.interfaces.nsIComponentRegistrar;
const nsIFactory                        = Components.interfaces.nsIFactory;
const nsIMsgAccount                     = Components.interfaces.nsIMsgAccount;
const nsIMsgAccountManager              = Components.interfaces.nsIMsgAccountManager;
const nsIMsgFolder                      = Components.interfaces.nsIMsgFolder;
const nsIPrefService                    = Components.interfaces.nsIPrefService;
const nsIRDFDataSource                  = Components.interfaces.nsIRDFDataSource;
const nsIRDFService                     = Components.interfaces.nsIRDFService
const nsISupports                       = Components.interfaces.nsISupports;
const nsISupportsArray                  = Components.interfaces.nsISupportsArray;
const nsISupportsString                 = Components.interfaces.nsISupportsString;

const NS_ERROR_FACTORY_REGISTER_AGAIN   = Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
const NS_ERROR_NO_AGGREGATION           = Components.results.NS_ERROR_NO_AGGREGATION;
const NS_ERROR_NO_INTERFACE             = Components.results.NS_ERROR_NO_INTERFACE;
const NS_ERROR_NOT_IMPLEMENTED          = Components.results.NS_ERROR_NOT_IMPLEMENTED;
const NS_ERROR_NOT_AVAILABLE            = Components.results.NS_ERROR_NOT_AVAILABLE;
const NS_ERROR_INVALID_ARG              = Components.results.NS_ERROR_INVALID_ARG;

const MSG_FOLDER_FLAG_TRASH             = 0x0100;

const QF_SEARCHROOT_ALL                 = 'msgaccounts:/';
const QF_SEARCHROOT_ACCOUNT             = 'msgaccount:/';
const QF_MAX_RESULTS                    = 25;

/***********************************************************
** qfMailNewsFoldersAutoComplete
***********************************************************/

function qfMailNewsFoldersAutoComplete () {
  this._accountManager = Components.classes['@mozilla.org/messenger/account-manager;1']
      .createInstance(nsIMsgAccountManager);

  /**
   * Holder for search parameters, like search root
   */
  this._searchParams =
  {
    searchRoot: 'msgaccounts:/'
  };

  /**
   * Since multiple accounts means multiple search
   *  roots, store these in an array
   */
  this._searchRoots = [];

  /**
   * Regex for search terms matching
   */
  this._regex = null;

  /**
   * The observer, either an nsIAutoCompleteListener or
   * an nsIAutoCompleteObserver, to report to
   */
  this._observer = null;

  /**
   * The search results, either an nsIAutoCompleteResults or
   * an nsIAutoCompleteResult. If this component is acting as
   * an nsIAutoCompleteSearch then this will be a qfAutoCompleteResult
   * implementing an nsIAutoCompleteResult. Otherwise it will be
   * an nsAutoCompleteResults
   */
  this._results = null;

  /**
   * The number of results currenly discovered. Used for limiting
   * the results to a reasonable number.
   * @see QF_MAX_RESULTS
   */
  this._resultsCount = 0;

  /**
   * Steal the quickfile.debug.enable preference
   */
  var prefs = Components.classes["@mozilla.org/preferences-service;1"]
          .getService(nsIPrefService)
          .getBranch("quickfile.");

  this._debug = false;
  if (prefs.getPrefType('debug.enable') == prefs.PREF_BOOL)
    this._debug = prefs.getBoolPref('debug.enable');
}

qfMailNewsFoldersAutoComplete.prototype =
{
  /**************************************************************
  ** nsISupports interface
  **************************************************************/

  /**
   *
   */
  QueryInterface: function _QueryInterface (aIID)
  {
    if (!aIID.equals(nsIAutoCompleteSession) &&
        !aIID.equals(nsIAutoCompleteSearch) &&
        !aIID.equals(nsICategoryManager) &&
        !aIID.equals(nsISupports))
      throw NS_ERROR_NO_INTERFACE;

    return this;
  },

  /**************************************************************
  ** nsIAutoCompleteSession interface
  **************************************************************/

  /**
   *
   */
  onStartLookup: function _onStartLookup (aSearchString, aPreviousSearchResult, aListener)
  {
    this._log("qfMailNewsFoldersAutoComplete::onStartLookup(" + aSearchString + ")");

    if (this._observer != aListener)
      this._observer = aListener;

    this._searchRoots = this._getSearchRoots(this._searchParams.searchRoot);
    this._generateRegex(aSearchString);

    this._initXPFEResults();
    this._results.searchString = aSearchString;

    this._search();

    var status = nsIAutoCompleteStatus.noMatch;
    if (this._results.items.Count())
      status = nsIAutoCompleteStatus.matchFound;

    this._observer.onAutoComplete(this._results, status);
  },

  /**
   *
   */
  onStopLookup: function _onStopLookup ()
  {
    this._log("qfMailNewsFoldersAutoComplete::onStopLookup()");
  },

  /**
   *
   */
  onAutoComplete: function _onAutoComplete (aSearchString, aPreviousSearchResult, aListener)
  {
    this._log("qfMailNewsFoldersAutoComplete::onAutoComplete(" + aSearchString + ")");
  },

  /**************************************************************
  ** nsIAutoCompleteSearch interface
  **************************************************************/

  /**
   *
   */
  startSearch: function _startSearch (aSearchString, aSearchParam, aPreviousResult,  aListener)
  {
    this._log("qfMailNewsFoldersAutoComplete::startSearch(" + aSearchString + ", " + aSearchParam + ")");

    if (this._observer != aListener)
      this._observer = aListener;

    // TODO: parse aSearchParam into this._searchParams
    this._searchParams.searchRoot = aSearchParam;
    this._searchRoots = this._getSearchRoots(this._searchParams.searchRoot);
    this._generateRegex(aSearchString);

    this._initToolkitResults();
    this._results.searchString = aSearchString;

    this._search();

    this._observer.onSearchResult(this, this._results);
  },

  /**
   *
   */
  stopSearch: function _stopSearch ()
  {
  },

  /**************************************************************
  ** nsICategoryManager interface
  **************************************************************/

  /**
   *
   */
  addCategoryEntry: function _addCategoryEntry (aCategory, aEntry, aValue, aPersist, aReplace)
  {
    this._log("qfMailNewsFoldersAutoComplete::addCategoryEntry(" + aCategory + ", " + aEntry + ", " + aValue + ", ..)");

    if (aCategory != 'quickfile')
      throw NS_ERROR_INVALID_ARG;

    // TODO: if aReplace = false && aEntry has value already
    //   throw NS_ERROR_INVALID_ARG

    switch (aEntry) {
      case 'searchRoot':
        this._searchParams[aEntry] = aValue;
        break;
      default:
        throw NS_ERROR_INVALID_ARG;
    }
  },

  deleteCategory: function _deleteCategory (aCategory)
  {
    throw NS_ERROR_NOT_IMPLEMENTED;
  },

  deleteCategoryEntry: function _deleteCategoryEntry (aCategory, aEntry, aPersist)
  {
    throw NS_ERROR_NOT_IMPLEMENTED;
  },

  enumerateCategories: function _enumerateCategories ()
  {
    // should return nsISimpleEnumerator
    throw NS_ERROR_NOT_IMPLEMENTED;
  },

  enumerateCategory: function _enumerateCategory (aCategory)
  {
    // should return nsISimpleEnumerator
    throw NS_ERROR_NOT_IMPLEMENTED;
  },

  getCategoryEntry: function _getCategoryEntry (aCategory, aEntry)
  {
    this._log("qfMailNewsFoldersAutoComplete::getCategoryEntry(" + aCategory + ", " + aEntry + ")");

    if (aCategory != 'quickfile')
      throw NS_ERROR_NOT_AVAILABLE;

    if (this._searchParams[aEntry] != undefined)
      return this._searchParams[aEntry];

    // if aCategory or aEntry doesn't exist
    throw NS_ERROR_NOT_AVAILABLE;
  },

  /**************************************************************
  ** PRIVATE methods
  **************************************************************/

  /**
   *
   */
  _log: function __log (aString)
  {
    if (this._debug)
      dump(aString + "\n");
  },

  /**
   *
   */
  _getSearchRoots: function __getSearchRoots (aSearchRootUri)
  {
    this._log("qfMailNewsFoldersAutoComplete::_getSearchRoots(" + aSearchRootUri + ")");

    // default to all accounts
    if (! aSearchRootUri || aSearchRootUri.length == 0)
      aSearchRootUri = QF_SEARCHROOT_ALL;

    var searchRoots = [];

    switch (aSearchRootUri) {
      case QF_SEARCHROOT_ALL:
        var accounts = this._accountManager.accounts.QueryInterface(nsISupportsArray);
        for (var i = 0; i < accounts.Count(); i++) {
          var account = accounts.QueryElementAt(i, nsIMsgAccount);
          if (! account)
            continue;

          var server = account.incomingServer;
          if (! server)
            continue;

          // don't search deferred accounts
          if ((server.rootFolder != server.rootMsgFolder) && ! server.isDeferredTo)
            continue;

          searchRoots.push(account.incomingServer.rootMsgFolder);
        }
        break;
      default:
        var msgfolder = this._getMsgFolderFromURI(aSearchRootUri);
        if (msgfolder)
          searchRoots.push(msgfolder);
        else
          return this._getSearchRoots(QF_SEARCHROOT_ALL);
        break;
    }

    return searchRoots;
  },

  /**
   *
   */
  _generateRegex: function __generateRegex (aSearchString)
  {
    this._log("qfMailNewsFoldersAutoComplete::_generateRegex(" + aSearchString + ")");

    var reString = aSearchString

    // escape meta-chars
    reString = reString.replace(/([\\^$?|*+(){}#\[\]])/g, '\\$1');

    // expand spaces
    reString = reString.split(/\s+/).join('.+')

    // case-insensitive
    var reOpts = 'i';

    this._regex = new RegExp(reString, reOpts);
  },

  /**
   * nsIAutoCompleteSession uses an nsIAutoCompleteResults to handle
   * passing back the data to the UI.
   */
  _initXPFEResults: function __initXPFEResults ()
  {
    this._log("qfMailNewsFoldersAutoComplete::_initXPFEResults()");

    this._results = Components.classes['@mozilla.org/autocomplete/results;1']
                              .createInstance(nsIAutoCompleteResults);
    this._resultsCount = 0;
  },

  /**
   * nsIAutoCompleteSearch uses an nsIAutoCompleteResult to handle
   * passing back the data to the UI. This is implemented here as
   * qfAutoCompleteResult.
   */
  _initToolkitResults: function __initToolkitResults ()
  {
    this._log("qfMailNewsFoldersAutoComplete::_initToolkitResults()");

    this._results = new qfAutoCompleteResult();
    this._resultsCount = 0;
  },

  /**
   *
   */
  _search: function __search ()
  {
    this._log("qfMailNewsFoldersAutoComplete::_search()");

    for (var i = 0; i < this._searchRoots.length; i++) {
      this._searchFolder(this._searchRoots[i], '');
    }
  },

  /**
   * Search a folder and its children for suitably qualified
   *  candidates.
   *
   * @param aFolder        the folder to search
   * @param aPrefixString  the full name of the parent folder
   */
  _searchFolder: function __searchFolder (aFolder, aPrefixString)
  {
    this._log("qfMailNewsFoldersAutoComplete::_searchFolder(" + aFolder.name + ")");

    // limit results
    if (this._resultsCount >= QF_MAX_RESULTS)
      return;

    var fullname = '';

    // don't include search root folders, unless there's more than one search root.
    // or we're searching everything, because that makes it impossible to determine
    // the label for the root folder..
    // ditto for searching the message account
    // NOTE: This is all because the UI needs to prepend the name of the search root
    //       folder to check we have the correct URI. If we're using msgaccounts:/ or
    //       msgaccount:/ then we find it hard to determine the correct label to 
    //       prepend, so avoid the problem here. Should be fixed
    // FIXME
    if (this._searchRoots.length > 1 || aFolder != this._searchRoots[0] ||
        this._searchParams.searchRoot == QF_SEARCHROOT_ALL ||
        this._searchParams.searchRoot == QF_SEARCHROOT_ACCOUNT) {
      fullname = (aPrefixString.length ? aPrefixString + '/' : '') + aFolder.name;
      if (this._qualifyCandidate(aFolder, fullname))
        this._appendResult(aFolder, fullname);
    }

    // don't descend into Trash
    if (aFolder.flags & MSG_FOLDER_FLAG_TRASH)
      return;

    if (aFolder.canCreateSubfolders && aFolder.hasSubFolders) {
      var subFoldersEnumerator = aFolder.GetSubFolders();
      try {
        subFoldersEnumerator.first();
        do {
          var subFolder = subFoldersEnumerator.currentItem().QueryInterface(nsIMsgFolder);
          this._searchFolder(subFolder, fullname);

          subFoldersEnumerator.next();
        }
        while (Components.lastResult == 0);
      }
      catch (ex) {
        // nsIEnumerator is a broken interface,
        // it actually throws an exception when next() is at the end
        //this.log("nsIEnumerator is broken! [" + ex + "]");
      }
    }
  },

  /**
   *
   */
  _qualifyCandidate: function __qualifyCandidate (aFolder, aFullName)
  {
    this._log("qfMailNewsFoldersAutoComplete::_qualifyCandidate(" + aFolder.name + ", " + aFullName + ")");

    if (! aFolder instanceof nsIMsgFolder || ! aFolder.canFileMessages ||
        ! this._regex.exec(aFullName))
      return false;

    return true;
  },

  /**
   *
   */
  _appendResult: function __appendResult (aFolder, aFullName)
  {
    this._log("qfMailNewsFoldersAutoComplete::_appendResult(" + aFolder.name + ", " + aFullName + ")");

    //TODO: don't switch each time, re-assign the appendResult method
    //      when creating the results, or on Qi
    if (this._results instanceof qfAutoCompleteResult)
      this._appendToolkitResult(aFolder, aFullName);
    else
      this._appendXPFEResult(aFolder, aFullName);

    this._resultsCount++;
  },

  /**
   *
   */
  _appendXPFEResult: function __appendXPFEResult (aFolder, aFullName)
  {
    this._log("qfMailNewsFoldersAutoComplete::_appendXPFEResult(" + aFolder.name + ", " + aFullName + ")");

    var result = Components.classes['@mozilla.org/autocomplete/item;1']
                           .createInstance(nsIAutoCompleteItem);

    result.value = aFullName;
    result.comment = aFolder.URI;

    this._results.items.AppendElement(result);
  },

  /**
   *
   */
  _appendToolkitResult: function __appendToolkitResult (aFolder, aFullName)
  {
    this._log("qfMailNewsFoldersAutoComplete::_appendToolkitResult(" + aFolder.name + ", " + aFullName + ")");

    var result = {
      value: aFullName,
      comment: aFolder.URI
    };

    this._results.appendResult(result);
  },

  /**
   * Ripped from mail/base/content/widgetgloue.js
   * @todo is this possible via XPCOM instead of RDF?
   */
  _getResourceFromURI: function __getResourceFromURI (aUri) {
    var RDF = Components.classes['@mozilla.org/rdf/rdf-service;1']
                        .getService()
                        .QueryInterface(nsIRDFService);

    return RDF.GetResource(aUri);
  },

  /**
   * Ripped from mail/base/content/widgetglue.js
   * @todo is this possible via XPCOM instead of RDF?
   */
  _getMsgFolderFromURI: function __getMsgFolderFromURI (aUri) {
    var msgfolder = null;
    try {
        var resource = this._getResourceFromURI(aUri);
        msgfolder = resource.QueryInterface(nsIMsgFolder);
    }
    catch (ex) {
        this._log("failed to get the folder resource for " + aUri);
    }
    return msgfolder;
  }
};

/**************************************************************
** qfAutoCompleteResult
**************************************************************/

const RESULT_IGNORED    = 1;
const RESULT_FAILURE    = 2;
const RESULT_NOMATCH    = 3;
const RESULT_SUCCESS    = 4;

/**
 *
 */
function qfAutoCompleteResult() {
  this._results = [];
}


qfAutoCompleteResult.prototype =
{
  /**************************************************************
  ** nsIAutoCompleteResult interface
  **************************************************************/

  defaultIndex: -1,
  errorDescription: null,
  matchCount: 0,
  searchResult: RESULT_NOMATCH,
  searchString: null,

  /**
   *
   */
  getCommentAt: function _getCommentAt (aIdx)
  {
    return this._results[aIdx].comment;
  },

  /**
   *
   */
  getStyleAt: function _getStyleAt (aIdx)
  {
    return this._results[aIdx].style;
  },

  /**
   *
   */
  getValueAt: function _getValueAt (aIdx)
  {
    return this._results[aIdx].value;
  },

  /**
   *
   */
  removeValueAt: function _removeValueAt (aIdx, aRemoveFromDb)
  {
    throw NS_ERROR_NOT_IMPLEMENTED;
  },

  /**************************************************************
  ** nsISupports interface
  **************************************************************/

  /**
   *
   */
  QueryInterface: function _QueryInterface (aIID)
  {
    if (!aIID.equals(nsIAutoCompleteResult) && !aIID.equals(nsISupports))
      throw NS_ERROR_NO_INTERFACE;

    return this;
  },

  /**************************************************************
  ** PUBLIC methods
  **************************************************************/

  /**
   *
   */
  appendResult: function _appendResult (aResult)
  {
    // results are useless without a label and a comment
    if (!(aResult && aResult.value && aResult.value.length) ||
        !(aResult.comment && aResult.comment.length))
      return;

    if (!(aResult.style && aResult.style.length))
      aResult.style = '';

    this._results.push(aResult);
    this.matchCount = this._results.length;
    this.searchResult = RESULT_SUCCESS;
  }
};

/**************************************************************
** Component registration etc
**************************************************************/

var qfMailNewsFoldersAutoCompleteFactory =
{
  createInstance: function _createInstance (aOuter, aIID)
  {
    if (aOuter != null)
      throw NS_ERROR_NO_AGGREGATION;

    return (new qfMailNewsFoldersAutoComplete()).QueryInterface(aIID);
  }
};

var qfMailNewsFoldersAutoCompleteModule =
{
  _firstTime: true,
  registerSelf: function _registerSelf (aCompMgr, aFileSpec, aLocation, aType)
  {
    // TODO: why defer here?
    if (this._firstTime) {
      this._firstTime = false;
      throw NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    aCompMgr = aCompMgr.QueryInterface(nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(XPFE_CLASS_ID, XPFE_CLASS_NAME, XPFE_CONTRACT_ID,
                                     aFileSpec, aLocation, aType);
    aCompMgr.registerFactoryLocation(TOOLKIT_CLASS_ID, TOOLKIT_CLASS_NAME, TOOLKIT_CONTRACT_ID,
                                     aFileSpec, aLocation, aType);
  },

  unregisterSelf: function _unregisterSelf (aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(XPFE_CLASS_ID, aLocation);
    aCompMgr.unregisterFactoryLocation(TOOLKIT_CLASS_ID, aLocation);
  },

  getClassObject: function _getClassObject (aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(nsIFactory))
      throw NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(XPFE_CLASS_ID) || aCID.equals(TOOLKIT_CLASS_ID))
      return qfMailNewsFoldersAutoCompleteFactory;

    throw NS_ERROR_NO_INTERFACE;
  },

  canUnload: function _canUnload (aCompMgr)
  {
    return true;
  }
};

function NSGetModule(aCompMgr, aFileSpec) {
  return qfMailNewsFoldersAutoCompleteModule;
}