/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Quickfile extension for Thunderbird.
 *
 * The Initial Developer of the Original Code is Paul Tomlin.
 * Portions created by the Initial Developer are Copyright (C) 2005, 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Paul Tomlin
 *  Mark Shuttleworth
 *
 * ***** END LICENSE BLOCK ***** */

/* $Id: install.js,v 1.1 2006/11/20 10:32:32 paultomlin Exp $ */

var contentFlag         = CONTENT | PROFILE_CHROME;
var localeFlag          = LOCALE | PROFILE_CHROME;
var skinFlag            = SKIN | PROFILE_CHROME;

var displayName         = "Quickfile";
var version             = "0.17.0.0011";

var name                = "quickfile";
var jarName             = "quickfile.jar";
var jarFolder           = "content/"

var error               = null;

var folder              = getFolder("Profile", "chrome");
var prefFolder          = getFolder(getFolder("Program", "defaults"), "pref");
var componentsFolder    = getFolder("Components");

var existsInApplication = File.exists(getFolder(getFolder("chrome"), jarName));
var existsInProfile     = File.exists(getFolder(folder, jarName));

var locales             = new Array( "en-US" );
var skins               = new Array( "classic" ); // "modern"
var prefs               = new Array( "preferences.js" );
var components          = new Array( "autocomplte.xpt", "qfMailNewsFoldersAutoComplete.js" );

// If the extension exists in the application folder or it doesn't exist
// in the profile folder and the user doesn't want it installed to the
// profile folder
if(existsInApplication ||
    (!existsInProfile &&
      !confirm( "Do you want to install the " + displayName +
                " extension into your profile folder?\n" +
                "(Cancel will install into the application folder)"))) {
    contentFlag = CONTENT | DELAYED_CHROME;
    folder      = getFolder("chrome");
    localeFlag  = LOCALE | DELAYED_CHROME;
    skinFlag    = SKIN | DELAYED_CHROME;
}

initInstall(displayName, name, version);
setPackageFolder(folder);
error = addFile(name, version, "chrome/" + jarName, folder, null);

// If adding the JAR file succeeded
if(error == SUCCESS) {
    folder = getFolder(folder, jarName);

    registerChrome(contentFlag, folder, jarFolder);
    for (var i = 0; i < locales.length; i++) {
        registerChrome(localeFlag, folder, "locale/" + locales[i] + "/");
    }

    for (var i = 0; i < skins.length; i++) {
        registerChrome(skinFlag, folder, "skin/" + skins[i] + "/");
    }

    for (var i = 0; i < prefs.length; i++) {
        addFile(name + " Defaults", version, "defaults/preferences/" + prefs[i],
            prefFolder, prefs[i], true);
    }

    for (var i = 0; i < components.length; i++) {
        addFile(name, version, "components/" + components[i],
            componentsFolder, components[i], true);
    }

    error = performInstall();

    // If the install failed
    if(error != SUCCESS && error != REBOOT_NEEDED) {
      displayError(error);
      cancelInstall(error);
    }
    else {
        alert("The installation of the " + displayName + " extension succeeded.");
    }
}
else {
  displayError(error);
  cancelInstall(error);
}

// Displays the error message to the user
function displayError(error) {
    // If the error code was -215
    if(error == READ_ONLY) {
        alert("The installation of " + displayName +
            " failed.\nOne of the files being overwritten is read-only.");
    }
    // If the error code was -235
    else if(error == INSUFFICIENT_DISK_SPACE) {
        alert("The installation of " + displayName +
            " failed.\nThere is insufficient disk space.");
    }
    // If the error code was -239
    else if(error == CHROME_REGISTRY_ERROR) {
        alert("The installation of " + displayName +
            " failed.\nChrome registration failed.");
    }
    else {
        alert("The installation of " + displayName +
            " failed.\nThe error code is: " + error);
    }
}